/*dots.c - generate patches of dots moving with arbitrary coherence
  Copyright (c) 1998 Matthew Belmonte

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  If you find this program useful, please send mail to Matthew Belmonte.
  <mkb4@Cornell.edu>.  If you base a publication on data processed by this
  program, please notify Matthew Belmonte and acknowledge him in your
  publication.
*/

#include <sys/types.h>
#include <sys/wait.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <ctype.h>
#include <errno.h>
#include <libgen.h>
#include <gl/image.h>
#define PI 3.1415926535897932384626433832795028841971693993751058209749445923078

int	*lbound, *rbound;	/*horizontal coordinates of boundaries*/
typedef struct {int x, y;} pair;
pair	*dot;			/*coordinates of all dots*/
char	*is_coherent,		/*tells whether each dot is moving nonrandomly*/
	*postproc,		/*name of postprocessor*/
	*postproc_program,	/*path name of postprocessor*/
	*postproc_filename_extension; /*extension to append to filenames*/
char snowp;			/*1 iff noncoherent dots randomly repositioned*/

static void configure_postprocessing()
  {
  postproc_program = getenv("VISTIM_PLUGIN");
  if(postproc_program != (char *)0)
    {
    postproc = pathfind(getenv("PATH"), postproc_program, "x"); /*no strdup() here, because this is the only use of pathfind()*/
    if(postproc == (char *)0)
      {
      fprintf(stderr, "%s: not found\n", postproc_program);
      exit(1);
      }
    postproc_filename_extension = getenv("VISTIM_EXT");
    }
  }

/*Allocate boundary lists.*/
static void alloc_mem(xsize)
int xsize;
  {
  lbound = (int *)calloc(xsize, sizeof(*lbound));
  rbound = (int *)calloc(xsize, sizeof(*rbound));
  if((lbound == (int *)0) || (rbound == (int *)0))
    {
    perror((char *)0);
    exit(errno);
    }
  }

/*Release dynamically allocated boundary lists.*/
static void free_mem()
  {
  free(lbound);
  free(rbound);
  }

/*Allocate dot data structures.*/
static void alloc_dot_mem(num_dots)
int num_dots;
  {
  dot = calloc(num_dots, sizeof(*dot));
  is_coherent = calloc(num_dots, sizeof(*is_coherent));
  if((dot == (pair *)0) || (is_coherent == (char *)0))
    {
    perror((char *)0);
    exit(errno);
    }
  }

/*Release dynamically allocated dot data structures.*/
static void free_dot_mem()
  {
  free(is_coherent);
  free(dot);
  }

/*Make the boundary arrays lbound[] and rbound[] map vertical coordinates to
  horizontal coordinates of a circular perimeter.  Return the number of pixels
  contained within the perimeter, including the area occupied by the perimeter
  itself.*/
static int set_boundaries(diameter)
int diameter;
  {
  register int y, area;
  double radius, root;
  radius = (diameter-1)/2.0;
  area = 0;
  for(y = 0; y != diameter; y++)
    {
    root = sqrt((double)(y*(diameter-y)));
    rbound[y] = (int)(0.5 + radius + root);
    if(rbound[y] >= diameter)
      rbound[y] = diameter-1;
    lbound[y] = (int)(0.5 + radius - root);
    if(lbound[y] < 0)
      lbound[y] = 0;
    area += 1+rbound[y]-lbound[y];
    }
  return(area);
  }

/*Choose 'num_moving' dots from the pool of 'num_dots' dots to move
  move coherently.  Set the is_coherent[] vector of flags accordingly.*/
static void select_moving_dots(r, num_dots)
double r;
int num_dots;
  {
  register int n, i;
  int num_coherent;
  num_coherent = (int)(0.499+r*num_dots);
  for(n = 0; n != num_dots; n++)
    is_coherent[n] = 0;
  for(n = 0; n != num_coherent; n++)
    {
    do
      i = random() % num_dots;
    while(is_coherent[i]);
    is_coherent[i] = 1;
    }
  }

/*set random coordinates for the ith dot*/
static void new_dot(i, h, v)
int	i,			/*index of this dot*/
	h, v;			/*bounds*/
  {
  dot[i].x = (int)(random() % h);
  dot[i].y = (int)(random() % v);
  }

/*Move the ith dot the given distance.  The direction of motion is the given
  theta if this dot is moving coherently (is_coherent[i]==1); otherwise the
  direction is randomly chosen.  Dots that fall out of the frame defined by h
  and v wrap around to the opposite side of the frame.*/
static void move_dot(i, h, v, r, theta)
int	i,			/*index of this dot*/
	h, v;			/*bounds*/
double	r,			/*distance of movement*/
	theta;			/*direction of movement (used iff is_coherent)*/
  {
/*If the "snow" mode is on, make noncoherent dots wink out and reappear
  elsewhere, rather than moving with the standard speed in a random direction.*/
  if(snowp && !is_coherent[i])
    {
    dot[i].x = (int)(random() % h);
    dot[i].y = (int)(random() % v);
    }
  else
    {
    if(!is_coherent[i])
      theta = (2.0*PI)*random()/((((unsigned long)1)<<31)-1L);
    dot[i].x = (dot[i].x + (int)rint(r*cos(theta))) % h;
    dot[i].y = (dot[i].y + (int)rint(r*sin(theta))) % v;
    if(dot[i].x < 0)
      dot[i].x += h;
    if(dot[i].y < 0)
      dot[i].y += v;
    }
  }

/*Generate an image with the current dot positions, keeping all out-of-bounds
  dots invisible.*/
static void generate_frame(h, v, num_dots)
int h, v, num_dots;
  {
  register int i;
  static int frame = 0;
  static unsigned short *imgbuf;
  char filename[80];
  IMAGE *img;
/*if this is the first call, allocate a buffer*/
  if(frame == 0)
    {
    imgbuf = calloc(h*v, sizeof(*imgbuf));
    if(imgbuf == (unsigned short *)0)
      {
      perror((char *)0);
      exit(errno);
      }
    }
/*zero the image*/
  for(i = 0; i != h*v; i++)
    imgbuf[i] = 0;
/*set all dots that fall within the boundaries of the visible area*/
  for(i = 0; i != num_dots; i++)
    if((dot[i].x >= lbound[dot[i].y]) && (dot[i].x <= rbound[dot[i].y]))
      imgbuf[(dot[i].y)*h + dot[i].x] = 0xff;
/*construct the filename for this frame*/
  sprintf(filename, "%04d.bw", frame++);
/*write the image*/
  img = iopen(filename, "w", RLE(1), 3, h, v, 1);
  for(i = 0; i != v; i++)
    putrow(img, imgbuf+i*h, i, 0);
  iclose(img);
/*if a postprocessor has been specified, execute it now*/
  if(postproc_program != (char *)0)
    {
    int status;
    sprintf(filename+60, "%%04d.%s", postproc_filename_extension);
    sprintf(filename+40, filename+60, frame-1);
    if(fork() == 0)
      {
      execl(postproc, postproc_program, filename, filename+40, (char *)0);
      perror(postproc_program);
      exit(errno);
      }
    if((wait(&status) <= 0) || !WIFEXITED(status))
      exit(1);
    if(WEXITSTATUS(status))
      exit(WEXITSTATUS(status));
    unlink(filename);
    }
  }

void main(argc, argv)
int argc;
char **argv;
  {
  register int d, frame;
  int	width, height,		/*dimensions of bounding box*/
	num_dots,		/*number of dots in the figure*/
	num_frames,		/*# of frames*/
	area;			/*imaged area (pixels)*/
  double dot_speed,		/*pixels per frame*/
	dot_direction,		/*degrees from right horizontal*/
	dot_correlation,	/*fraction of dots that move non-randomly*/
	dot_density;		/*interpretation depends on dot_density_spec*/
  char dot_density_spec;
  if(argc >= 8)
    {
/*width of square frame*/
    height = width = atoi(argv[1]);
/*speed and direction of motion*/
    dot_speed = atof(argv[2]);
    dot_direction = atof(argv[3]);
    if(dot_speed < 0)			/*negative speed inverts direction*/
      {
      dot_direction += 180.0;
      dot_speed *= -1.0;
      }
    dot_direction -= 360.0*floor(dot_direction/360.0); /*standardise range*/
    dot_direction *= PI/180.0;		/*convert to radians*/
/*correlation of dot motions*/
    dot_correlation = atof(argv[4]);
    if((dot_correlation < 0.0) || (dot_correlation > 100.0))
      {
      fprintf(stderr, "correlation must be between 0% and 100%\n");
      exit(1);
      }
    dot_correlation /= 100.0;		/*convert from % to fraction*/
/*number of dots in image*/
    dot_density_spec=toupper(argv[5][0]);
    dot_density = atof(argv[6]);
    switch(dot_density_spec)
      {
      case 'D':
	alloc_mem(width);
	area = set_boundaries(width);
      /*specified as density; multiply by area of bounding box*/
	num_dots = (int)(0.5+dot_density*width*height);
	break;
      case 'N':
	alloc_mem(width);
	area = set_boundaries(width);
      /*specified as absolute number of dots; correct for area of box*/
	num_dots = (int)((0.5+dot_density)*(width*height)/area);
	break;
      default:
	fprintf(stderr, "dot-density specifier must be (D)ots-per-pixel or (N)umber-of-dots\n");
	exit(1);
      }
    num_frames = atoi(argv[7]);
    if(num_frames < 1)
      {
      fprintf(stderr, "must have at least one frame\n");
      exit(1);
      }
    snowp = ((argc >= 9) && (argv[8][0]=='s') && (argv[8][1]=='n') && (argv[8][2]=='o') && (argv[8][3]=='w') && (argv[8][4]=='\0'));
    }
  else
    {
    fprintf(stderr, "usage: %s <width> <speed> <direction> <correlation> <density spec> <density> <# of frames> [snow]\n", *argv);
    fprintf(stderr, "PARAMETER\t\tTYPE\tUNITS\n");
    fprintf(stderr, "width\t\tinteger\tpixels\n");
    fprintf(stderr, "speed\t\tfloat\tpixels per frame\n");
    fprintf(stderr, "direction\tfloat\tdegrees from right horizontal\n");
    fprintf(stderr, "correlation\tfloat\tpercent\n");
    fprintf(stderr, "density spec\tchar\t(D)ots-per-pixel | (N)umber-of-dots\n");
    fprintf(stderr, "dot density\tfloat\t[see density spec above]\n");
    fprintf(stderr, "# of frames\tinteger\tdimensionless\n");
    exit(1);
    }
  configure_postprocessing();
  alloc_dot_mem(num_dots);
/*initial positions*/
  for(frame = 0; frame != num_dots; frame++)
    new_dot(frame, width, height);
  generate_frame(width, height, num_dots);
/*successive positions*/
  for(frame = 1; frame < num_frames; frame++)
    {
    select_moving_dots(dot_correlation, num_dots);
    for(d = 0; d != num_dots; d++)
      move_dot(d, width, height, dot_speed, dot_direction);
    generate_frame(width, height, num_dots);
    }
  free_dot_mem();
  free_mem();
  exit(0);
  }
