/*gabor.c - generate a Gabor patch
  by Matthew Belmonte
  Copyright (c) 1998 Howard Hughes Medical Institute

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  If you find this program useful, please send mail to Matthew Belmonte.
  <mkb4@Cornell.edu>.  If you base a publication on data processed by this
  program, please notify Matthew Belmonte and acknowledge him in your
  publication.
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <gl/image.h>
#define PI 3.1415926535897932384626433832795028841971693993751058209749445923078
#define DMAX 1280	/*maximum image diameter*/

/*the Normal density function N(x), with mean m and standard deviation s*/
double normal(x, m, s)
double x, m, s;
  {
  double e;
  e = (x-m)/s;
  return(exp(-e*e/2.0)/sqrt(2.0*PI*s*s));
  }

double gabor_xscale, gabor_yscale; /*factors for input & output of normal fn.*/

/*Zero in on the deviation for which the normal function equals one part in the
  entire greyscale range for the given number of bits - for example, if
  num_bits is 8, then the range is 255, so the function finds the deviation for
  which the normal function equals 1/255.  (That deviation turns out the be
  3.0405.)*/
void zero_in(num_bits)
int num_bits;
  {
  register double x,	/*current guess as to the x-value at fade-out*/
	x_lb,		/*lower bound on the x-value at fade-out*/
	x_ub;		/*upper bound on the x-value at fade-out*/

/*ratio of the greyscale range to the maximum output of the Normal distribution;
  multiplying the output of the Normal distribution by this scaling value will
  produce a value between zero and the maximum greyscale value, inclusive*/

  gabor_yscale = ((double)((1<<num_bits)-1))/normal(0.0, 0.0, 1.0);

  x_lb = 0.0;		/*initial lower bound on x is zero*/
  x_ub = 5.0;		/*N(x,0,1) where x>5 is tiny and may as well be 0*/
  do
    {
    x = (x_lb+x_ub)/2.0;

  /*if the scaled value is less than one - i.e. if the amplitude of the Normal
    function at this point relative to its maximum is less than one part in the
    entire greyscale range, make this the new upper bound*/
    if(gabor_yscale*normal(x, 0.0, 1.0) < 1.0)

      x_ub = x;
    else
      x_lb = x;
    }while(x_ub > x_lb+0.00000000000001);
  gabor_xscale = x;
  }

/*Express x, the eccentricity of the current pixel, as a fraction of the radius
  of the Gabor patch.  (Note that since the patch sits in a rectangular frame,
  this fraction may be greater than one, for positions near the corners of that
  frame.)  Multiply this by 'gabor_xscale', a global that stretches the Normal
  distribution on its horizontal axis s.t. the resulting Gabor patch fades'
  gently to black at the edges of the frame.  Feed this scaled distance to the
  standard (zero-mean, unit-variance) Normal function.  Scale the output of
  that function by 'gabor_yscale', a global that stretches the Normal
  distribution on its vertical axis, similarly to the aforementioned horizontal
  stretch.*/
unsigned scaled_normal(x, radius)
double x, radius;
  {
  return((unsigned int)(gabor_yscale*normal(x*gabor_xscale/radius, 0.0, 1.0)));
  }

/*Draw a Gabor patch of the specified diameter, and place it in the given
  file.*/
void draw_gabor(d, filename)
int d;		/*diameter*/
char *filename;
  {
  register int x, y;
  register double r;
  IMAGE *img;
  unsigned short row[DMAX];
  img = iopen(filename, "w", RLE(1), 3, d, d, 1);
  if(d & 1)
    {
  /*diameter is odd; handle the meridian as a special case*/
    r = (double)(d/2);
    row[d/2] = scaled_normal(0.0, r);
    for(x = 1; x != (d+1)/2; x++)
      row[d/2 - x] = row[d/2 + x] = scaled_normal((double)x, r);
    putrow(img, row, d/2, 0);
  /*compute non-meridian pixels*/
    for(y = 1; y != (d+1)/2; y++)
      {
      row[d/2] = scaled_normal((double)y, r);
      for(x = 1; x != (d+1)/2; x++)
	row[d/2 - x] = row[d/2 + x] = scaled_normal(sqrt((double)(x*x+y*y)), r);
      putrow(img, row, d/2-y, 0);
      putrow(img, row, d/2+y, 0);
      }
    }
  else
  /*diameter is even; all pixels are symmetric with other pixels*/
    {
    r = d/2 - 0.5;
    for(y = 0; y != d/2; y++)
      {
      for(x = 0; x!= d/2; x++)
	row[d/2-x-1] = row[d/2+x] = scaled_normal(sqrt((double)(x*x+y*y+x+y+0.5)), r);
      putrow(img, row, d/2-y-1, 0);
      putrow(img, row, d/2+y, 0);
      }
    }
  iclose(img);
  }

void main(argc, argv)
int argc;
char **argv;
  {
  int gabor_diameter;
  if((argc != 3) || ((gabor_diameter=atoi(argv[1])) <= 2))
    {
    fprintf(stderr, "usage: %s <diameter> <file name>\n", *argv);
    exit(1);
    }
  zero_in(8);
  draw_gabor(gabor_diameter, argv[2]);
  exit(0);
  }
