/*gabor_movie.c - make a movie of a sine-wave grating in a Gabor envelope.
  by Matthew Belmonte
  Copyright (c) 1998 Howard Hughes Medical Institute

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  If you find this program useful, please send mail to Matthew Belmonte.
  <mkb4@Cornell.edu>.  If you base a publication on data processed by this
  program, please notify Matthew Belmonte and acknowledge him in your
  publication.
*/

#include <sys/types.h>
#include <sys/wait.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <libgen.h>

static char gabor_program[] = "gabor",
	grating_program[] = "grating",
	mult_program[] = "multimg";
static char gabor_tempfile[] = "gabor.bw",
	grating_tempfile[] = "grating.bw";

char *gabor, *grating, *mult, *postproc, *postproc_program, *postproc_filename_extension;

static char *find_helper(name, path)
char *name, *path;
  {
  char *pathname;
  pathname = pathfind(path, name, "x");
  if(pathname == (char *)0)
    {
    fprintf(stderr, "%s: not found\n", name);
    exit(1);
    }
  return(strdup(pathname));
  }

static void configure_helpers()
  {
  char *path;
/*get search path*/
  path = getenv("PATH");
  if(path == (char *)0)
    {
    fprintf(stderr, "error: PATH environment variable is null.\n");
    exit(1);
    }
/*search for helper applications*/
  gabor = find_helper(gabor_program, path);
  grating = find_helper(grating_program, path);
  mult = find_helper(mult_program, path);
  postproc_program = getenv("VISTIM_PLUGIN");
  if(postproc_program != (char *)0)
    postproc = find_helper(postproc_program, path);
  postproc_filename_extension = getenv("VISTIM_EXT");
  }

void main(argc, argv)
int argc;
char **argv;
  {
  register int frame;	/*frame count*/
  int status;		/*exit status of child process*/
  double init_phase,	/*initial phase at centre*/
	tf,		/*temporal frequency*/
	frate,		/*monitor frame rate*/
	fper,		/*frame period = 1/frate*/
	dur,		/*duration of movie (seconds)*/
	time;		/*time at midpoint of current frame*/
  char *ampl_buf;	/*negative amplitude*/
  char buf[80];		/*scratch space for execl parameters*/
  if(argc != 9)
    {
    fprintf(stderr, "usage: %s <width> <sper> <phase> <ori> <tf> <ampl> <frate> <dur>\n", *argv);
    fprintf(stderr, "PARAM\tTYPE\tUNITS\n");
    fprintf(stderr, "width\tinteger\tpixels\n");
    fprintf(stderr, "sper\tfloat\tpixels\n");
    fprintf(stderr, "phase\tfloat\tdegrees\n");
    fprintf(stderr, "ori\tfloat\tdegrees from vertical\n");
    fprintf(stderr, "tf\tfloat\tHertz\n");
    fprintf(stderr, "ampl\tfloat\t-1..1 (negative values invert phase)\n");
    fprintf(stderr, "frate\tfloat\tHertz\n");
    fprintf(stderr, "dur\tfloat\tseconds\n");
    fprintf(stderr, "\nexample: %s 256 32 90 45 10 1 100 0.1\n", *argv);
    exit(1);
    }
  init_phase = atof(argv[3]);
  tf = atof(argv[5]);
/*if amplitude is negative, make ampl_buf point to it, and invert the phase*/
  if(argv[6][0] == '-')
    {
    ampl_buf = argv[6];
    init_phase += 180.0;
    }
/*if amplitude is positive, make a copy of it with a leading '-'*/
  else
    {
    frame = 0;
    buf[40] = '-';
    do
      buf[41+frame] = argv[6][frame];
      while(argv[6][frame++] != '\0');
    ampl_buf = buf+40;
    }
  frate = atof(argv[7]);
  fper = 1.0/frate;
  dur = atof(argv[8]);
  configure_helpers();
/*draw the Gabor patch*/
  if(fork() == 0)
    {
    execl(gabor, gabor_program, argv[1], gabor_tempfile, (char *)0);
    perror(gabor);
    exit(errno);
    }
  if((wait(&status) <= 0) || !WIFEXITED(status))
    exit(1);
  if(WEXITSTATUS(status))
    exit(WEXITSTATUS(status));
/*draw frames:*/
  for(frame = 0, time = 0.5*fper; time < dur; frame++, time += fper)
    {
  /*draw grating at current phase*/
    sprintf(buf, "%f", init_phase + (frame/frate)*tf*360.0);
    if(fork() == 0)
      {
      execl(grating, grating_program, argv[1], argv[2], buf, argv[4], grating_tempfile, (char *)0);
      perror(grating_program);
      exit(errno);
      }
    if((wait(&status) <= 0) || !WIFEXITED(status))
      exit(1);
    if(WEXITSTATUS(status))
      exit(WEXITSTATUS(status));
  /*multiply by the Gabor envelope*/
    sprintf(buf, "%03d.bw", frame);
    if(fork() == 0)
      {
      execl(mult, mult_program, gabor_tempfile, "0", "1", grating_tempfile, ampl_buf, ampl_buf+1, buf, "-1", "1", (char *)0);
      perror(mult_program);
      exit(errno);
      }
    if((wait(&status) <= 0) || !WIFEXITED(status))
      exit(1);
    if(WEXITSTATUS(status))
      exit(WEXITSTATUS(status));
  /*if a postprocessor has been specified, feed it the output, & delete orig.*/
    if(postproc_program != (char *)0)
      {
      if(fork() == 0)
	{
	sprintf(buf+60, "%%03d.%s", postproc_filename_extension);
	sprintf(buf+40, buf+60, frame);
	execl(postproc, postproc_program, buf, buf+40, (char *)0);
	perror(postproc_program);
	exit(errno);
	}
      if((wait(&status) <= 0) || !WIFEXITED(status))
	exit(1);
      if(WEXITSTATUS(status))
	exit(WEXITSTATUS(status));
      unlink(buf);
      }
    }
  unlink(grating_tempfile);
  unlink(gabor_tempfile);
  exit(0);
  }
