/*multimg.c - linearly map grey values to reals in a series of images, multiply
  the mapped values, and map the resultant back to grey values.
  by Matthew Belmonte
  Copyright (c) 1998 Howard Hughes Medical Institute

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  If you find this program useful, please send mail to Matthew Belmonte.
  <mkb4@Cornell.edu>.  If you base a publication on data processed by this
  program, please notify Matthew Belmonte and acknowledge him in your
  publication.
*/

#include <stdio.h>
#include <stdlib.h>
#include <gl/image.h>
#define MAX_FILES 24

/*Multiply the vector 'row', with 0 mapped to 'low' and 255 mapped to 'high',
  into the vector 'vec'.  Both vectors are of length 'w'.*/
void pointwise(row, vec, w, low, high)
unsigned short *row;
double *vec;
int w;
double low, high;
  {
  register int x;
  for(x = 0; x != w; x++)
    vec[x] *= low+row[x]*(high-low)/255.0;
  }

void main(argc, argv)
int argc;
char **argv;
  {
  register int x, y, file;
  int w, h, n;				/*width, height, # of images*/
  unsigned short *row;
  double *vec;
  IMAGE *img[1+MAX_FILES];
  double low[1+MAX_FILES], high[1+MAX_FILES];
  if((argc < 7) || (argc%3 != 1))
    {
    fprintf(stderr, "usage: %s <image.bw> <low map> <high map> ...\n", *argv);
    fprintf(stderr, "example: %s gabor.bw 0 1 grating.bw -1 1 gabor_grating.bw -1 1\n", *argv);
    exit(1);
    }
  if(argc/3-1 > MAX_FILES)
    {
    fprintf(stderr, "recompile %s with MAX_FILES=%d\n", *argv, argc/3-1);
    exit(1);
    }
/*open inputs*/
  img[0] = iopen(argv[1], "r");
  if(img[0] == (IMAGE *)0)
    exit(1);
  sscanf(argv[2], "%lf", low);
  sscanf(argv[3], "%lf", high);
  w = img[0]->xsize;
  h = img[0]->ysize;
  n = argc/3-1;
  for(file = 1; file < n; file++)
    {
    img[file] = iopen(argv[1+3*file], "r");
  /*check dimensions*/
    if((img[file]->xsize != w) || (img[file]->ysize != h))
      {
      fprintf(stderr, "dimensions %d x %d don't match %d x %d; ", img[file]->xsize, img[file]->ysize, w, h);
      iclose(img[file]);
      img[file] = (IMAGE *)0;
      }
  /*if open failed or dimensions didn't match, back out of all open files*/
    if(img[file] == (IMAGE *)0)
      {
      fprintf(stderr, "couldn't open %s\n", argv[1+3*file]);
      while(file--)
	iclose(img[file]);
      exit(1);
      }
  /*tabulate floating-point mappings for low (0) and high (255) ends of range*/
    sscanf(argv[2+3*file], "%lf", low+file);
    sscanf(argv[3+3*file], "%lf", high+file);
    }
/*initialise row vectors*/
  row = (unsigned short *)calloc(w, sizeof(*row));
  vec = (double *)calloc(w, sizeof(*vec));
/*open output*/
  img[n] = iopen(argv[argc-3], "w", RLE(1), 3, w, h, 1);
  if(img[n] == (IMAGE *)0)
    {
    fprintf(stderr, "can't open output file %s\n", argv[argc-3]);
    exit(1);
    }
  sscanf(argv[argc-2], "%lf", low+n);
  sscanf(argv[argc-1], "%lf", high+n);
/*for each row...*/
  for(y = 0; y != h; y++)
    {
  /*initialise to identity...*/
    for(x = 0; x != w; x++)
      vec[x] = 1.0;
  /*multiply in...*/
    for(file = 0; file != n; file++)
      {
      getrow(img[file], row, y, 0);
      pointwise(row, vec, w, low[file], high[file]);
      }
  /*multiply out...*/
    for(x = 0; x != w; x++)
      {
      vec[x] = 0.5+255.0*(vec[x] - low[n])/(high[n] - low[n]);
      if(vec[x] > 255.0) vec[x] = 255.0; else if(vec[x] < 0.0) vec[x] = 0.0;
      row[x] = (unsigned int)(vec[x]);
      }
  /*output*/
    putrow(img[n], row, y, 0);
    }
/*close all files*/
  for(file = n; file >= 0; file--)
    iclose(img[file]);
/*free memory*/
  free(vec);
  free(row);
  exit(0);
  }
